import 'dart:async';

import 'package:msg/src/msg.dart';

typedef Effect = void Function();

abstract class Cmd<Model> {
  Cmd();

  factory Cmd.none() => _CmdNone();
  factory Cmd.ofMsg(FutureOr<Msg<Model>> msg) => _CmdMsg(msg);
  factory Cmd.ofEffect(Effect effect) => _CmdEffect(effect);
  factory Cmd.batch(List<Cmd<Model>> cmds) => _CmdBatch(cmds);
}

// MARK: -- None

abstract class CmdNone<Model> extends Cmd<Model> {}

class _CmdNone<Model> implements CmdNone<Model> {}

// MARK: -- Of Msg

abstract class CmdMsg<Model> extends Cmd<Model> {
  FutureOr<Msg<Model>> get msg;
}

class _CmdMsg<Model> implements CmdMsg<Model> {
  final FutureOr<Msg<Model>> msg;

  _CmdMsg(this.msg);
}

// MARK: -- Of Effect

abstract class CmdEffect<Model> extends Cmd<Model> {
  Effect get effect;
}

class _CmdEffect<Model> implements CmdEffect<Model>  {
  final Effect effect;

  _CmdEffect(this.effect);
}

// MARK: -- Batch

abstract class CmdBatch<Model> extends Cmd<Model> {
  List<Cmd<Model>> get cmds;
}

class _CmdBatch<Model> implements CmdBatch<Model> {
  final List<Cmd<Model>> cmds;

  _CmdBatch(this.cmds);
}
