import 'dart:async';

import 'package:msg/src/cmd.dart';
import 'package:msg/src/msg.dart';
import 'package:msg/src/store_base.dart';

class Store<Model> implements StoreBase<Model> {
  final _msgController = StreamController<Msg<Model>>();
  final _modelController = StreamController<Model>.broadcast();

  late final StreamSubscription<void> _msgSubscription;
  late final List<StreamSubscription<void>> _externalSubscriptions;

  Model _model;

  Store({
    required Model initialModel,
    Cmd<Model>? initialCmd,
    List<Stream<Msg<Model>>> Function(Model model)? subscriptions,
  }) : _model = initialModel {
    void handleCmd(Cmd<Model> cmd) {
      if (cmd is CmdMsg<Model>) {
        final msg = cmd.msg;
        if (msg is Future<Msg<Model>>) {
          msg.then((msg) => _msgController.add(msg));
        } else {
          _msgController.add(msg);
        }
      }
      if (cmd is CmdEffect<Model>) {
        cmd.effect();
      }
      if (cmd is CmdBatch<Model>) {
        cmd.cmds.forEach((cmd) => handleCmd(cmd));
      }
    }

    _msgSubscription = _msgController.stream.listen((msg) {
      final old = _model;
      if (msg.update != null) {
        _model = msg.update!(_model);
        if (_model != old) {
          _modelController.add(_model);
        }
      }
      if (msg.cmd != null) {
        handleCmd(msg.cmd!(old));
      }
    });

    _externalSubscriptions = subscriptions == null
        ? []
        : subscriptions(initialModel)
            .map((stream) => stream.listen((msg) => _msgController.add(msg)))
            .toList();

    if (initialCmd != null) {
      handleCmd(initialCmd);
    }
  }

  @override
  void add(Msg<Model> msg) => _msgController.add(msg);

  @override
  void close() {
    _msgSubscription.cancel();
    _externalSubscriptions.forEach((subscription) => subscription.cancel());
    _msgController.close();
    _modelController.close();
  }

  Model get model => _model;

  Stream<Model> get stream => _modelController.stream;
}
